#!/usr/bin/env python
# -*- coding: iso-8859-1 -*-

import msnp
import random
import string
import time




##########################################
#                                        #
#  Classe de base de tous les listeners  #
#                                        #
##########################################

class MsnChatListener(msnp.ChatCallbacks):
    """Cette classe est la classe de base des gestionnaires de jeux
    qui répondent au chat."""

    welcomeMessage = "Bonjour!  Je ne suis pas Martin, mais plutôt "       + \
                     "un petit programme qu'il a fait pour te divertir.  " + \
                     "Si tu veux jouer avec moi, tape quelque chose et "   + \
                     "on va s'amuser!"
    stopWord = "stop"
    listeners = []
    listenerIterator = 0


    def stopWordInMessage(self, message):
        """Cette fonction vérifie que le mot d'arrêt a été reçu."""
        for word in self.convert(message).split():
            if word == self.stopWord:
                return True
        return False


    def convert(self, message, charset='utf-8'):
        """Cette fonction convertit la chaîne reçu en une chaîne
        utilisable par le programme."""
        return message.decode(charset).encode('iso-8859-1')


    def end(self, firstTime=False):
        callbacks = self.listeners[self.listenerIterator]()
        callbacks.listenerIterator = (self.listenerIterator + 1) % len(self.listeners)

        self.chat.callbacks = callbacks
        callbacks.chat = self.chat
        self.chat = None
        if firstTime:
            callbacks.sendWelcomeMessage()
        else:
            callbacks.sendMessage("Je viens de changer de comportement...  Amuse-toi!")


    def sendMessage(self, message, charset='utf-8'):
        """Cette fonction envoie un message en s'assurent que
        l'encodage est correct."""
        self.chat.send_message(message.decode('iso-8859-1').encode(charset))


    def sendWelcomeMessage(self):
        """Cette fonction envoie le message de bienvenue."""
        self.sendMessage(self.welcomeMessage)




#############################################################
#                                                           #
#  Classe qui ignore les entrées et change de gestionnaire  #
#                                                           #
#############################################################

class IgnoreMsnChatListener(MsnChatListener):
    """Cette classe est utilisée pour ignorer la première entrée qui
    cause la conversation."""

    def message_received(self, passport_id, display_name, text, charset):
        self.end(True)




##################################
#                                #
#  Classe du jeu des anagrammes  #
#                                #
##################################

class AnagramMsnChatListener(MsnChatListener):
    """Cette classe implante le jeu des anagrammes."""


    def __init__(self):
        self.anagram = None
        self.answer  = None


    def scramble(self, message):
        message = list(message)
        random.shuffle(message)
        return ''.join(message)


    def message_received(self, passport_id, display_name, text, charset):
        if self.anagram == None:
            # On prend un mot dans le texte qu'il a écrit: le plus long
            words = self.convert(text).split()
            words.sort(cmp=lambda x,y : len(y) - len(x))
            if len(words) > 0:
                self.answer  = words[0]
                self.anagram = self.scramble(words[0])
            self.sendMessage('Voici un anagramme: "' + self.anagram + \
                             '".  Bonne chance!')
        else:
            # Est-ce que le mot a été deviné?
            words = self.convert(text).split()

            answerFound   = False
            stopWordFound = False
            for word in words:
                if word == self.answer:
                    answerFound = True
                if word == MsnChatListener.stopWord:
                    stopWordFound = True

            if answerFound:
                self.sendMessage('Bravo!  Tu as trouvé!')
                self.anagram = None
                self.answer  = None
            else:
                if not stopWordFound:
                    self.sendMessage("Désolé, ce n'est pas le mot...  " + \
                                     'Tu peux réessayer ou taper "'     + \
                                     MsnChatListener.stopWord + '" '    + \
                                     'si tu es tanné!')
                    self.sendMessage('(Le mot est "' + self.anagram + '")')
                else:
                    self.sendMessage("Merci d'avoir joué!")
                    self.sendMessage('(La réponse était "' + self.answer + '")')
                    self.anagram = None
                    self.answer  = None
                    self.end()

# On enregistre la classe comme listener valide
MsnChatListener.listeners.append(AnagramMsnChatListener)




###############################################################
#                                                             #
#  Classe qui change la présentation du message et le répète  #
#                                                             #
###############################################################

class PrettyMsnChatListener(MsnChatListener):
    """Cette classe implante une répétition du message embelli."""


    def __init__(self):
        self.answererIterator = 0


    def same(self, text, charset):
        return self.convert(text)


    def upperLower(self, text, charset):
        message = self.convert(text)
        returnValue = []
        i = 0
        while i < len(message):
            if (i % 2 == 0):
                function = string.lower
            else:
                function = string.upper
            returnValue.append(function(message[i]))
            i = i + 1
        return ''.join(returnValue)


    def star(self, text, charset):
        message = self.convert(text)
        returnValue = []
        i = 0
        while i < len(message):
            returnValue.append(message[i])
            returnValue.append('*')
            i = i + 1
        return '*' + ''.join(returnValue)


    def colored(self, text, charset):
        # Ce code est copié/collé de chat.py
        import email
        import email.Message
        mime_message = email.Message.Message()
        mime_message['MIME-Version'] = '1.0'
        mime_message['Content-Type'] = 'text/plain; charset=' + charset
        # On change le style du message
        mime_message['X-MMS-IM-Format'] = 'EF=BI CO=0000ff'
        mime_message.set_payload(text)
        self.chat._Chat__send_mime_message(mime_message, 'N')
        # Retourner none spécifie que le message est déjà envoyé
        return None


    converters = [
        same ,
        upperLower ,
        star ,
        colored
        ]


    def message_received(self, passport_id, display_name, text, charset):
        if (self.stopWordInMessage(text)):
            self.sendMessage("Merci de t'être amusé!")
            self.end()
        else:
            function = self.converters[self.answererIterator]
            self.answererIterator = (self.answererIterator + 1) % len(self.converters)
            response = function(self, text, charset)
            if response != None:
                self.sendMessage(response)
            self.sendMessage('(Tu peux taper "' + self.stopWord + \
                             '" si tu es tanné!)')

# On enregistre la classe comme listener valide
MsnChatListener.listeners.append(PrettyMsnChatListener)




########################################
#                                      #
#  Classe du jeu de deviner un numéro  #
#                                      #
########################################

class GuessMsnChatListener(MsnChatListener):
    """Cette classe implante le jeu de deviner un numéro."""


    maxValue = 100
    minValue = 0


    def __init__(self):
        self.number   = None


    def message_received(self, passport_id, display_name, text, charset):
        if self.number == None:
            # On choisit un nombre à faire deviner
            self.minGuess = self.minValue
            self.maxGuess = self.maxValue
            self.number   = random.randint(self.minValue, self.maxValue)

            self.sendMessage("J'ai choisi un nombre entre " + \
                             str(self.minValue) + " et " + \
                             str(self.maxValue) + ".  Devine-le!")
        else:
            if (self.stopWordInMessage(text)):
                self.sendMessage("Merci d'avoir joué!")
                self.sendMessage("(La réponse était " + str(self.number) + ")")
                self.end()
            else:
                # On essaie de trouver un nombre dans le message
                number = None
                for word in self.convert(text).split():
                    try:
                        number = int(word)
                    except ValueError:
                        pass

                needToExplainStop = True
                if number == None:
                    self.sendMessage("Tu dois entrer un nombre!")
                elif not number > self.minGuess:
                    self.sendMessage("Ce n'est pas super intelligent de " + \
                                     "choisir un nombre plus petit ou égal à " + \
                                     str(self.minGuess) + "...")
                elif not number < self.maxGuess:
                    self.sendMessage("Ce n'est pas super intelligent de " + \
                                     "choisir un nombre plus grand ou égal à que " + \
                                     str(self.maxGuess) + "...")
                elif number < self.number:
                    self.sendMessage("Bel essai, mais trop bas!")
                    self.minGuess = number
                elif number > self.number:
                    self.sendMessage("Bel essai, mais trop haut!")
                    self.maxGuess = number
                else:
                    self.sendMessage("Bravo!  Tu as trouvé le numéro!  " + \
                                     "On recommence!")
                    needToExplainStop = False
                    self.number = None
                    # On rappelle la fonction
                    self.message_received(passport_id, display_name, None, None)

                if (needToExplainStop):
                    self.sendMessage('(Tu peux taper "' + self.stopWord + \
                                     '" si tu es tanné!)')

# On enregistre la classe comme listener valide
MsnChatListener.listeners.append(GuessMsnChatListener)




######################################################################
######################################################################
######################################################################




#####################################################
#                                                   #
#  Classe de réception générale des énévements MSN  #
#                                                   #
#####################################################

class MsnListener(msnp.SessionCallbacks):
    def state_changed(self, state):
        if state == msnp.States.ONLINE:
            print 'MSNBot est maintenant en ligne'
    def friend_list_updated(self, friend_list):
        print 'Réception de la liste de contacts:'
        for friend in friend_list.get_friends():
            print 'passport_id: ' + friend.get_passport_id() + \
                  ' , display_name: ' + friend.get_display_name()
    def chat_started(self, chat):
        callbacks = IgnoreMsnChatListener()
        callbacks.chat = chat
        chat.callbacks = callbacks


msn = msnp.Session(MsnListener())
msn.login('address@hotmail.com', 'password')
msn.sync_friend_list()

while True:
    msn.process(chats = True)
    time.sleep(1)



# Notes:
# - J'ai modifié session.py pour tenir compte des messsages "NOT" qui
#   m'empêchait de me connecter avec mon compte



# error.py -- Error class(es)
#
# Copyright (C) 2003 Manish Jethani (manish_jethani AT yahoo.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

class Error:
    """Generic exception type"""
    def __init__(self, code, message):
        self.code = code
        self.message = message
    def __str__(self):
        return str(self.code) + ':' + self.message

class HttpError(Error):
    """Error returned from HTTP server"""
    def __init__(self, code, message, http_status, http_reason):
        Error.__init__(self, code, message)
        self.http_status = http_status
        self.http_reason = http_reason
    def __str__(self):
        return Error.__str__(self) \
        + '[%s:%s]' % (str(self.http_status), self.http_reason)

# vim: set ts=4 sw=4 et tw=79 :




# net.py -- Connection, HttpProxyConnection classes
#
# Copyright (C) 2003 Manish Jethani (manish_jethani AT yahoo.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import socket
import httplib

from string import split, join

class Connection:  # generic tcp connection wrapper
    def __init__(self, server):
        self.socket = None
        self.server = server

    def establish(self):
        if self.socket == None:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            server = (self.server[0], int(self.server[1]))
            s.connect(server)
            self.socket = s
        return self.socket

    def send_data(self, buf):
        return self.socket.send(buf)

    def receive_data(self, bufsize):
        return self.socket.recv(bufsize)

    def send_data_all(self, buf):
        total = len(buf)
        sent = 0
        while sent < total:
            sent = sent + self.send_data(buf[sent:])
        return sent

    def send_data_line(self, line):
        # print "C:" + line #XXX
        return self.send_data_all(line) + self.send_data_all('\r\n')

    def receive_data_line(self):
        cnt = 0
        buf = ''
        while 1:
            in_byte = self.receive_data(1)
            if in_byte == '':
                return None
            if in_byte == '\r':
                cnt = 1
            elif in_byte == '\n' and cnt == 1:
                cnt = 2
            else:
                cnt = 0
            buf = buf + in_byte
            if cnt == 2:
                # print "S:" + buf #XXX
                return buf

    def break_(self):
        self.socket.shutdown(2)
        self.socket.close()
        self.socket = None

class HttpProxyConnection(Connection):  # http tunnelling
    def __init__(self, server, proxy):
        Connection.__init__(self, server)
        self.proxy = proxy

    def establish(self):
        tmp = self.server
        self.server = self.proxy
        try:
            Connection.establish(self)
        finally:
            self.server = tmp

        connect_str = 'CONNECT ' + self.server[0] \
            + ':' + str(self.server[1]) \
            + ' HTTP/1.0\r\n'
        self.send_data_all(connect_str)
        self.send_data_all('User-Agent: msnp.py\r\n')
        self.send_data_all('Host: ' + self.server[0] + '\r\n')
        self.send_data_all('\r\n')

        status = -1
        while 1:
            buf = self.receive_data_line()
            if status == -1:
                resp = split(buf, ' ', 2)
                if len(resp) > 1:
                    status = int(resp[1])
                else:
                    status = 0
            if buf == '\r\n':
                break

        if status != 200:
            self.socket = None
        return self.socket

class HTTPSConnection(httplib.HTTPSConnection):
    # httplib.HTTPSConnection with HTTP proxy support
    def __init__(self, host, port = None, key_file = None, cert_file = None,
                 strict = None, http_proxy = None):
        httplib.HTTPSConnection.__init__(self, host, port, key_file, cert_file,
                                         strict)
        self.http_proxy = http_proxy

    def connect(self):
        if self.http_proxy:
            conn = HttpProxyConnection((self.host, self.port), self.http_proxy)
            conn.establish()
            sock = conn.socket
            ssl = socket.ssl(sock, self.key_file, self.cert_file)
            self.sock = httplib.FakeSocket(sock, ssl)
        else:
            httplib.HTTPSConnection.connect(self)

# vim: set ts=4 sw=4 et tw=79 :




# session.py -- Session, SessionCallbacks classes
#
# Copyright (C) 2003 Manish Jethani (manish_jethani AT yahoo.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import select
import md5

from string import split, join
from binascii import hexlify, unhexlify
from time import time

from protocol import States, Lists, PrivacyModes
from error import Error, HttpError
from friend import Group, Friend, FriendList
from net import Connection, HttpProxyConnection
from command import Command, Msg, Png, Qry
from codec import url_codec

import protocol
import chat

class _Session:  # common base for Session and Chat
    def __init__(self, callbacks):
        self.callbacks = callbacks
        self.transaction_id = 0
        self.http_proxy = None
        self.conn = None
        self.send_queue = []

    def _connect(self, server):
        conn = None
        if self.http_proxy:
            conn = HttpProxyConnection(server, self.http_proxy)
        else:
            conn = Connection(server)
        conn.establish()
        return conn

    def _increment_transaction_id(self):
        self.transaction_id = self.transaction_id + 1
        return self.transaction_id

    def _send_cmd(self, cmd, conn):
        conn.send_data_line(str(cmd))
        self._increment_transaction_id()

    def _receive_cmd(self, conn):
        buf = conn.receive_data_line()
        if buf == None:  # connection closed
            raise Error(1, 'Connection closed.')
        cmd = Command()
        cmd.parse(buf)
        return cmd

    def _sync_command(self, cmd, conn):
        # synchronous command (receive response immediately)
        self._send_cmd(cmd, conn)
        return self._receive_cmd(conn)

    def _async_command(self, cmd):
        self.send_queue.append(cmd)
        self._increment_transaction_id()

class SessionCallbacks:  # callback interface
    """Callback interface for MSN instant messaging session

    To receive notification on various protocol events, the client must
    implement some or all of the methods in this callback interface.
    """

    def ping(self):
        """Ping received from server"""

    def state_changed(self, state):
        """User's presence state has changed

        Keyword arguments:
            state -- any of the msnp.States members
        """

    def friend_online(self, state, passport_id, display_name):
        """Friend is online

        Keyword arguments:
            state -- any of the msnp.States members
            passport_id -- string representing friend's passport ID
            display_name -- friend's display name
        """

    def friend_offline(self, passport_id):
        """Friend is offline

        Keyword arguments:
            passport_id -- string representing friend's passport ID
        """

    def friend_list_updated(self, friend_list):
        """Friend list has been updated

        Keyword arguments:
            friend_list -- same as msnp.Session.friend_list
        """

    def logged_out(self):
        """User has been logged out"""

    def group_added(self, id, name):
        """Group has been added

        Keyword arguments:
            id -- group ID
            name -- name of group
        """

    def group_removed(self, id):
        """Group has been removed

        Keyword arguments:
            id -- group ID
        """

    def group_renamed(self, id, name):
        """Group has been renamed

        Keyword arguments:
            id -- group ID
            name -- new name of group
        """

    def friend_added(self, list_, passport_id, display_name, group_id = -1):
        """Friend has been added

        If list_ is msnp.Lists.REVERSE, it means that the user has been added
        to someone's list.  In that case, the passport_id and display_name
        parameters contain information about that someone.

        Keyword arguments:
            list_ -- type of list (allow, block, etc.)
            passport_id -- string representing friend's passport ID
            display_name -- friend's display name
            group_id -- group ID of group to which friend has been added
        """

    def friend_removed(self, list_, passport_id, group_id = -1):
        """Friend has been removed

        Keyword arguments:
            list_ -- type of list (allow, block, etc.)
            passport_id -- string representing friend's passport ID
            group_id -- group ID of group from which friend has been removed
        """

    def display_name_changed(self, display_name):
        """Display name changed

        Keyword arguments:
            display_name -- user's new display name
        """

    def display_name_received(self, passport_id, display_name):
        """Display name received

        Keyword arguments:
            passport_id -- string representing friend's passport ID
            display_name -- friend's display name
        """

    def chat_started(self, chat):
        """Chat started

        Keyword arguments:
            chat -- Chat instance representing new chat started
        """

class Session(_Session):
    """MSN instant messaging session

    To get into an instant messaging session, an instance of msnp.Session must
    be created.  The session can be started by calling the login method.  After
    logging in, the process method must be called periodically to process the
    server's commands.
    """

    class __ChatRequest:
        def __init__(self, invitee):
            self.invitee = invitee

    def __init__(self, callbacks = None, dispatch_server = None):
        """Constructor for msnp.Session

        Keyword arguments:
            callbacks -- callback interface
            dispatch_server -- dispatch server host, port
        """

        if callbacks == None:
            callbacks = SessionCallbacks()
        _Session.__init__(self, callbacks)

        if dispatch_server == None:
            self.dispatch_server = ('messenger.hotmail.com', 1863)

        self.logged_in = 0
        self.passport_id = None
        self.display_name = None
        self.chat_requests = {}
        self.friend_list = FriendList()
        self.active_chats = {}

    def __get_twn_ticket(self, twn_string, username, password):
        from net import HTTPSConnection
        from urllib import urlencode
        debuglevel = 0

        # step 1: get address of login server
        con = HTTPSConnection('nexus.passport.com',
            http_proxy = self.http_proxy)
        con.set_debuglevel(debuglevel)
        con.request('GET', '/rdr/pprdr.asp')
        res = con.getresponse()
        con.close()
        if res.status != 200:
            raise HttpError(0, 'Bad response from passport nexus server.',
                res.status, res.reason)
        hdr = res.getheader('PassportURLs')
        url = {}
        for u in hdr.split(','):
            k, v = u.split('=')
            url[k] = v
        dalogin = url['DALogin'].split('/', 1)

        # step 2: get "ticket" to notification server
        while True:
            con = HTTPSConnection(dalogin[0], http_proxy = self.http_proxy)
            con.set_debuglevel(debuglevel)
            auth = 'Passport1.4 OrgVerb=GET,%s,%s,%s,%s' \
                % (urlencode({'OrgURL': 'http://messenger.msn.com'}),
                    urlencode({'sign-in': username}),
                    urlencode({'pwd': password}),
                    twn_string)
            con.request('GET', '/%s' % (dalogin[1]), '',
                {'Authorization': auth})
            res = con.getresponse()
            con.close()
            if res.status != 200:
                raise HttpError(0, 'Bad response from login server.',
                    res.status, res.reason) # XXX handle redirection?
            else:
                break
        hdr = res.getheader('Authentication-Info') or \
              res.getheader('WWW-Authenticate')
        hdr = hdr[len('Passport1.4 '):]
        auth = {}
        for u in hdr.split(','):
            k, v = u.split('=', 1)
            if v[0] == '\'' and v[-1] == '\'':
                v = v[1:-1]
            auth[k] = v
        ticket = auth['from-PP']

        return ticket
        # TODO code cleanup

    def __handshake(self, server, username, password):
        conn = self._connect(server)
        try:
            ver = Command('VER', self.transaction_id, ('MSNP8', 'CVR0'))
            resp = self._sync_command(ver, conn)
            if resp.cmd != 'VER' or resp.args[0] == '0':
                raise Error(0, 'Bad response for VER command.')

            cvr = Command('CVR', self.transaction_id,
                ('0x0409', 'win', '4.10', 'i386', 'MSNMSGR', '6.0.0602',
                'MSMSGS ', username))
            resp = self._sync_command(cvr, conn)
            if resp.cmd != 'CVR':
                raise Error(0, 'Bad response for CVR command.')

            usr = Command('USR', self.transaction_id, ('TWN', 'I', username))
            resp = self._sync_command(usr, conn)
            if resp.cmd != 'USR' and resp.cmd != 'XFR':
                raise Error(0, 'Bad response for USR command.')

            # for dispatch server, response is ver, cvr, xfr; for notification
            # server, it is ver, cvr, usr (or same as dispatch server, in some
            # cases)

            if resp.cmd == 'XFR':
                return split(resp.args[1], ':', 1)
            elif resp.cmd == 'USR':
                twn_string = resp.args[2]

                ticket = self.__get_twn_ticket(twn_string, username, password)

                usr = Command('USR', self.transaction_id, ('TWN', 'S', ticket))
                resp = self._sync_command(usr, conn)
                if resp.cmd != 'USR':
                    raise Error(int(resp.cmd), protocol.errors[resp.cmd])
                elif resp.args[0] != 'OK':
                    raise Error(0, 'Bad response for USR command.')

                self.passport_id = resp.args[1]
                self.display_name = url_codec.decode(resp.args[2])
                self.logged_in = 1
        finally:
            if not self.logged_in:
                conn.break_()
            else:
                self.conn = conn

    def process(self, chats = False):
        """Process events

        Keyword arguments:
            chats -- whether or not to call msnp.Chat.process for all active
                chat sessions

        This method must be called periodically, preferably in the client
        application's main loop.
        """
        while self.logged_in:
            fd = self.conn.socket.fileno()
            r = select.select([fd], [], [], 0)
            if len(r[0]) > 0:
                buf = self.conn.receive_data_line()
                self.__process_command_buf(buf)
            elif len(self.send_queue) > 0:
                cmd = self.send_queue.pop(0)
                cmd.send(self.conn)
            else:
                break
        if chats:
            self.__process_active_chats()

    def __process_active_chats(self):
        [chat_.process() for chat_ in self.active_chats.values()]

    def __process_command_buf(self, buf):
        cmd = buf[:3]
        if cmd == 'MSG':
            self.__process_msg(buf)
        elif cmd == 'QNG':
            self.__process_qng(buf)
        elif cmd == 'OUT':
            self.__process_out(buf)
        elif cmd == 'RNG':
            self.__process_rng(buf)
        ### AJOUT PAR MARTIN BISSON
        elif cmd == 'NOT':
            # On ignore la prochaine ligne
            self.conn.receive_data_line()
        ### FIN DE L'AJOUT
        else:
            c = Command()
            c.parse(buf)
            if c.cmd == 'CHG':
                self.__process_chg(c)
            elif c.cmd == 'ILN':
                self.__process_iln(c)
            elif c.cmd == 'NLN':
                self.__process_nln(c)
            elif c.cmd == 'FLN':
                self.__process_fln(c)
            elif c.cmd == 'CHL':
                self.__process_chl(c)
            elif c.cmd == 'LSG':
                self.__process_lsg(c)
            elif c.cmd == 'LST':
                self.__process_lst(c)
            elif c.cmd == 'SYN':
                self.__process_syn(c)
            elif c.cmd == 'XFR':
                self.__process_xfr(c)
            elif c.cmd == 'BLP':
                self.__process_blp(c)
            elif c.cmd == 'GTC':
                self.__process_gtc(c)
            elif c.cmd == 'ADG':
                self.__process_adg(c)
            elif c.cmd == 'RMG':
                self.__process_rmg(c)
            elif c.cmd == 'REG':
                self.__process_reg(c)
            elif c.cmd == 'ADD':
                self.__process_add(c)
            elif c.cmd == 'REM':
                self.__process_rem(c)
            elif c.cmd == 'REA':
                self.__process_rea(c)
            elif c.cmd == '218':
                pass
            # TODO error handling

    def __process_msg(self, buf):
        msg = Msg()
        msg.parse(buf)
        msg.receive(self.conn)
        # discard NS messages for now

    def __process_qng(self, buf):
        self.callbacks.ping()

    def __process_out(self, buf):
        self.conn.break_()
        self.conn = None
        self.logged_in = 0
        self.callbacks.logged_out()

    def __process_rng(self, buf):
        cmdline = split(buf)
        session_id = cmdline[1]
        sb = split(cmdline[2], ':')
        server = (sb[0], int(sb[1]))
        hash = cmdline[4]
        passport_id = cmdline[5]
        display_name = url_codec.decode(cmdline[6])
        try:
            chat_ = chat.Chat(self, server, hash, passport_id,
                display_name, session_id)
        except Error, e:
            if e.code == 1:  # connection closed
                return
            raise e
        self.active_chats[chat_.session_id] = chat_
        self.callbacks.chat_started(chat_)

    def __process_chg(self, command):
        self.callbacks.state_changed(command.args[0])

    def __process_iln(self, command):
        state = command.args[0]
        passport_id = command.args[1]
        display_name = url_codec.decode(command.args[2])
        friend = self.friend_list.get_friend(passport_id)
        if friend != None:
            friend.state = state
            friend.display_name = display_name
            self.__friend_list_updated()
        else:  # usu. immed. after login
            self.friend_list.temp_iln[passport_id] = state
        self.callbacks.friend_online(state, passport_id, display_name)

    def __process_nln(self, command):
        state = command.args[0]
        passport_id = command.args[1]
        display_name = url_codec.decode(command.args[2])
        friend = self.friend_list.get_friend(passport_id)
        if friend != None:
            friend.display_name = display_name
            friend.state = state
            self.__friend_list_updated()
        self.callbacks.friend_online(state, passport_id, display_name)

    def __process_fln(self, command):
        passport_id = command.args[0]
        friend = self.friend_list.get_friend(passport_id)
        if friend != None:
            friend.state = States.OFFLINE
            self.__friend_list_updated()
        self.callbacks.friend_offline(passport_id)

    def __process_chl(self, command):
        qry = Qry(self.transaction_id, command.args[0])
        self._async_command(qry)

    def __process_lsg(self, command):
        id = int(command.args[0])
        name = url_codec.decode(command.args[1])

        group = Group(id, name)
        self.friend_list.groups[id] = group

        self.__friend_list_updated()

    def __process_lst(self, command):
        from protocol import list_flags

        passport_id  = command.args[0]
        display_name = url_codec.decode(command.args[1])
        list_        = int(command.args[2])
        group_id     = []

        if list_ & list_flags[Lists.FORWARD]:
            group_id = [int(i) for i in split(command.args[3], ',')]

        groups = None
        if len(group_id):
            groups = [self.friend_list.groups[g_id] for g_id in group_id]

        friend = Friend(passport_id, display_name, groups = groups)
        for f in list_flags.keys():
            if list_ & list_flags[f]:
                self.friend_list.lists[f][passport_id] = friend

        if self.friend_list.temp_iln.has_key(passport_id):
            friend.state = self.friend_list.temp_iln[passport_id]

        self.__friend_list_updated()

    def __process_syn(self, command):
        ver = int(command.args[0])
        self.friend_list.ver = ver
        self.__friend_list_updated()

    def __process_xfr(self, command):
        sb = split(command.args[1], ':')
        server = (sb[0], int(sb[1]))
        cr = self.chat_requests[command.trn]
        invitee = cr.invitee
        chat_ = chat.Chat(self, server, command.args[3], self.passport_id,
            self.display_name, None, invitee)
        self.active_chats[chat_.session_id] = chat_
        self.callbacks.chat_started(chat_)

    def __process_blp(self, command):
        privacy_mode = command.args[0]
        self.friend_list.privacy_mode = privacy_mode
        self.__friend_list_updated()

    def __process_gtc(self, command):
        notify_on_add = command.args[0] == 'A'
        self.friend_list.notify_on_add_ = notify_on_add
        self.__friend_list_updated()

    def __process_adg(self, command):
        ver = int(command.args[0])
        name = url_codec.decode(command.args[1])
        id = int(command.args[2])
        self.friend_list.ver = ver
        self.friend_list.groups[id] = Group(id, name)
        self.__friend_list_updated()
        self.callbacks.group_added(id, name)

    def __process_rmg(self, command):
        ver = int(command.args[0])
        id = int(command.args[1])
        self.friend_list.ver = ver
        if self.friend_list.groups.has_key(id):
            del self.friend_list.groups[id]
        self.__friend_list_updated()
        self.callbacks.group_removed(id)

    def __process_reg(self, command):
        ver = int(command.args[0])
        id = int(command.args[1])
        name = url_codec.decode(command.args[2])
        self.friend_list.ver = ver
        if self.friend_list.groups.has_key(id):
            self.friend_list.groups[id].name = name
        self.__friend_list_updated()
        self.callbacks.group_renamed(id, name)

    def __process_add(self, command):
        list_ = command.args[0]
        ver = int(command.args[1])
        passport_id = command.args[2]
        display_name = url_codec.decode(command.args[3])
        group = None
        if list_ == Lists.FORWARD:
            group = self.friend_list.groups[int(command.args[4])]

        self.friend_list.ver = ver
 
        friend = self.friend_list.get_friend(passport_id, list_)
        if friend != None:
            friend.add_to_group(group)
        else:
            if group != None:
                friend = Friend(passport_id, passport_id, (group))
            else:
                friend = Friend(passport_id, passport_id)
            self.friend_list.lists[list_][passport_id] = friend

        self.__friend_list_updated()

        if group != None:
            self.callbacks.friend_added(list_, passport_id, display_name,
                group.get_id())
        else:
            self.callbacks.friend_added(list_, passport_id, display_name)

    def __process_rem(self, command):
        list_ = command.args[0]
        ver = int(command.args[1])
        passport_id = command.args[2]
        group = None
        if list_ == Lists.FORWARD:
            group = self.friend_list.groups[int(command.args[3])]

        self.friend_list.ver = ver

        friend = self.friend_list.get_friend(passport_id, list_)
        if friend != None: # this shouldn't be None, unless friend_list stale
            if group != None:
                friend.remove_from_group(group)
            if len(friend.get_groups()) == 0:
                del self.friend_list.lists[list_][passport_id]

        self.__friend_list_updated()
        if group != None:
            self.callbacks.friend_removed(list_, passport_id, group.get_id())
        else:
            self.callbacks.friend_removed(list_, passport_id)

    def __process_rea(self, command):
        ver = int(command.args[0])
        passport_id = command.args[1]
        display_name = url_codec.decode(command.args[2])

        if passport_id == self.passport_id:
            self.display_name = display_name
            self.callbacks.display_name_changed(display_name)
        else:
            self.callbacks.display_name_received(passport_id, display_name)

    def __friend_list_updated(self):
        self.friend_list.updated = time()
        self.callbacks.friend_list_updated(self.friend_list)

    def login(self, username, password, initial_state = States.ONLINE):
        """Login to MSN server

        Keyword arguments:
            username -- username
            password -- password
            initial_state -- initial state (default msnp.States.ONLINE)
        """
        if self.logged_in:
            return
        server = self.dispatch_server
        while not self.logged_in:
            server = self.__handshake(server, username, password)
        self.change_state(initial_state)

    def ping(self):
        """Ping server"""
        if not self.logged_in:
            return
        self._async_command(Png())
        self.process()

    def logout(self):
        """Logout from server"""
        if not self.logged_in:
            return
        [chat_.leave() for chat_ in self.active_chats.values()]
        self.process()
        self.conn.break_()
        self.conn = None
        self.logged_in = 0

    def change_state(self, state):
        """Change user's state

        Keyword arguments:
            state -- new state (see msnp.States)
        """
        if not self.logged_in:
            return
        chg = Command('CHG', self.transaction_id, (state,))
        self._async_command(chg)
        self.process()

    def sync_friend_list(self, ver = -1):
        """Synchronise friend list by getting new copy from server

        The friend list is updated asynchronously.
        msnp.SessionCallbacks.friend_list_updated will be called repeatedly
        after a call to this method.  The client may want to set a timer
        instead, and check for updates to the friend list using the
        msnp.FriendList.last_updated method.

        Keyword arguments:
            ver -- friend list version
        """
        if not self.logged_in:
            return
        self.friend_list.dirty = False
        if ver == -1:
            ver = self.friend_list.ver
        syn = Command('SYN', self.transaction_id, (str(ver),))
        self._async_command(syn)
        self.process()

    def request_list(self, list_ = Lists.FORWARD):
        """Request a list from the server

        Keyword arguments:
            list_ -- type of list to request (see msnp.Lists)
        """
        if not self.logged_in:
            return
        lst = Command('LST', self.transaction_id, (list_,))
        self._async_command(lst)
        self.process()

    def request_groups(self):
        """Request groups from server"""
        if not self.logged_in:
            return
        lsg = Command('LSG', self.transaction_id, ())
        self._async_command(lsg)
        self.process()

    def change_privacy_mode(self, privacy_mode):
        """Change privacy mode

        Keyword arguments:
            privacy_mode -- new privacy mode (see msnp.PrivacyModes)
        """
        if not self.logged_in:
            return
        blp = Command('BLP', self.transaction_id, (privacy_mode,))
        self._async_command(blp)
        self.process()

    def notify_on_add(self, notify):
        """Change setting for being notified on being added"""
        if not self.logged_in:
            return
        setting = 'N'
        if notify:
            setting = 'A'
        gtc = Command('GTC', self.transaction_id, (setting,))
        self._async_command(gtc)
        self.process()

    def add_group(self, name):
        """Add a group
        
        Keyword arguments:
            name -- name of new group
        """
        if not self.logged_in:
            return
        adg = Command('ADG', self.transaction_id,
            (url_codec.encode(name), '0'))
        self._async_command(adg)
        self.process()

    def remove_group(self, id):
        """Remove a group

        Keyword arguments:
            id -- group ID
        """
        if not self.logged_in:
            return
        rmg = Command('RMG', self.transaction_id, (str(id),))
        self._async_command(rmg)
        self.process()

    def rename_group(self, id, name):
        """Rename a group

        Keyword arguments:
            id -- group ID
            name -- new name of group
        """
        if not self.logged_in:
            return
        reg = Command('REG', self.transaction_id,
            (str(id), url_codec.encode(name), '0'))
        self._async_command(reg)
        self.process()

    def add_friend(self, list_, passport_id, group_id = 0):
        """Add a friend

        Keyword arguments:
            list_ -- type of list (allow, block, etc.)
            passport_id -- string representing friend's passport ID
            group_id -- group ID of group to which friend is being added
        """
        add = None
        if list_ == Lists.FORWARD:
            add = Command('ADD', self.transaction_id,
                (list_, passport_id, passport_id, str(group_id)))
        else:
            add = Command('ADD', self.transaction_id,
                (list_, passport_id, passport_id))
        self._async_command(add)
        self.process()

    def remove_friend(self, list_, passport_id, group_id = 0):
        """Remove a friend

        Keyword arguments:
            list_ -- type of list (allow, block, etc.)
            passport_id -- string representing friend's passport ID
            group_id -- group ID of group from which friend is being removed
        """
        rem = None
        if list_ == Lists.FORWARD:
            rem = Command('REM', self.transaction_id,
                (list_, passport_id, str(group_id)))
        else:
            rem = Command('REM', self.transaction_id,
                (list_, passport_id))
        self._async_command(rem)
        self.process()

    def change_display_name(self, display_name):
        """Change user's display name

        Keyword arguments:
            display_name -- user's new display name
        """
        if not self.logged_in:
            return
        rea = Command('REA', self.transaction_id,
            (self.passport_id, url_codec.encode(display_name)))
        self._async_command(rea)
        self.process()

    def request_display_name(self, passport_id):
        """Request display name of a friend

        Keyword arguments:
            passport_id -- string representing friend's passport ID
        """
        if not self.logged_in:
            return
        rea = Command('REA', self.transaction_id,
            (passport_id, url_codec.encode('MJ++')))
        self._async_command(rea)
        self.process()

    def start_chat(self, invitee):
        """Start a chat

        Keyword arguments:
            invitee -- friend invited for chat
        """
        if not self.logged_in:
            return
        xfr = Command('XFR', self.transaction_id, ('SB',))
        self._async_command(xfr)
        self.chat_requests[xfr.trn] = Session.__ChatRequest(invitee)
        self.process()

# vim: set ts=4 sw=4 et tw=79 :




# codec.py -- UrlCodec class, url_codec global
#
# Copyright (C) 2003 Manish Jethani (manish_jethani AT yahoo.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from string import split, join
from binascii import hexlify, unhexlify

class UrlCodec:  # url-encoding (rfc 1738)
    def encode(self, str):
        if str == None:
            return None
        buf = []
        for c in str:
            h = hexlify(c)
            i = ord(c)
            if 0x00 <= i <= 0x20 or i == 0x7f or 0x80 <= i <= 0xff:
                buf.append('%')
                buf.append(h)
            else:
                buf.append(c)
        return join(buf, '')

    def decode(self, str):
        if str == None:
            return None
        buf = []
        p = hex = None
        for c in str:
            if c == '%':
                p = '%'
            elif p != None:
                (hex, p) = (c, None)
            elif hex != None:
                hex = hex + c
                buf.append(unhexlify(hex))
                hex = None
            else:
                buf.append(c)
        return join(buf, '')
url_codec = UrlCodec()

# vim: set ts=4 sw=4 et tw=79 :




# msnp.py -- An implementation of the MSN Messenger Protocol
#
# Version 0.4 (for Python 2.3)
#
# Copyright (C) 2003 Manish Jethani (manish_jethani AT yahoo.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""An implementation of the MSN Messenger Protocol

msnp is a library, written in object-oriented Python, for accessing the MSN
instant messaging service.

To use this module for instant messaging, an instance of the msnp.Session class
must be created.  The msnp.SessionCallbacks interface must be implemented by
the client.  For example:

    import msnp

    class MyCallbacks(msnp.SessionCallbacks):
        def state_changed(self, state):
            print 'New state:', state

    im = msnp.Session(MyCallbacks())
    im.login('gill_bates@hotmail.com', 'microshaft')

If the login succeeds, the state_changed method of the MyCallbacks instance
will be called, and 'New state: NLN' will get printed.

For more information, visit:
http://msnp.sourceforge.net/
"""

# Note about documentation:  All public objects (classes, methods, globals,
# etc.) have been documented with Python docstrings; private code has been
# sprinkled with terse comments.

from error import Error
from friend import Group, Friend, FriendList

from chat import Chat, ChatCallbacks
from session import Session, SessionCallbacks

from protocol import States, Lists, PrivacyModes

import protocol

__all__ = [
    'Error',
    'Group', 'Friend', 'FriendList',
    'Chat', 'ChatCallbacks',
    'Session', 'SessionCallbacks',
    'States', 'Lists', 'PrivacyModes',
]

# vim: set ts=4 sw=4 et tw=79 :




# protocol.py -- States, Lists, PrivacyModes classes, and errors global
#
# Copyright (C) 2003 Manish Jethani (manish_jethani AT yahoo.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

class States:
    """Presence states"""
    ONLINE = 'NLN'
    OFFLINE = 'FLN'
    INVISIBLE = 'HDN'
    BUSY = 'BSY'
    IDLE = 'IDL'
    BE_RIGHT_BACK = 'BRB'
    AWAY = 'AWY'
    ON_THE_PHONE = 'PHN'
    OUT_TO_LUNCH = 'LUN'

class Lists:
    """Types of friend lists"""
    FORWARD = 'FL'
    REVERSE = 'RL'
    ALLOW = 'AL'
    BLOCK = 'BL'

class PrivacyModes:
    """Privacy modes"""
    ALLOW = 'AL'
    BLOCK = 'BL'

list_flags = {
    Lists.FORWARD : 1,
    Lists.ALLOW   : 2,
    Lists.BLOCK   : 4,
    Lists.REVERSE : 8,
}

errors = {
    '200': 'Syntax error',
    '201': 'Invalid parameter',
    '205': 'Invalid user',
    '206': 'Domain name missing',
    '207': 'Already logged in',
    '208': 'Invalid username',
    '209': 'Invalid fusername',
    '210': 'User list full',
    '215': 'User already there',
    '216': 'User already on list',
    '217': 'User not online',
    '218': 'Already in mode',
    '219': 'User is in the opposite list',
    '219': 'User is in the opposite list',
    '231': 'Tried to add a contact to a group that doesn\'t exist',
    '280': 'Switchboard failed',
    '281': 'Transfer to switchboard failed',

    '300': 'Required field missing',
    '302': 'Not logged in',

    '500': 'Internal server error',
    '501': 'Database server error',
    '510': 'File operation failed',
    '520': 'Memory allocation failed',
    '540': 'Wrong CHL value sent to server',

    '600': 'Server is busy',
    '601': 'Server is unavaliable',
    '602': 'Peer nameserver is down',
    '603': 'Database connection failed',
    '604': 'Server is going down',

    '707': 'Could not create connection',
    '710': 'CVR parameters either unknown or not allowed',
    '711': 'Write is blocking',
    '712': 'Session is overloaded',
    '713': 'Too many active users',
    '714': 'Too many sessions',
    '715': 'Not expected',
    '717': 'Bad friend file',

    '911': 'Authentication failed',
    '913': 'Not allowed when offline',
    '920': 'Not accepting new users',
    '924': 'Passport account not yet verified',
}

# vim: set ts=4 sw=4 et tw=79 :




# friend.py -- Group, Friend, FriendList classes
#
# Copyright (C) 2003 Manish Jethani (manish_jethani AT yahoo.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from protocol import States, Lists, PrivacyModes
from time import time

class Group:
    """A group in the friend list"""
    def __init__(self, id, name, friends = None):
        self.id = id
        self.name = name
        if friends == None:
            friends = {}
        self.friends = friends

    def size(self):
        """Return number of friends in this group"""
        return len(self.friends)

    def get_id(self):
        """Return ID of this group"""
        return self.id

    def get_name(self):
        """Return name of this group"""
        return self.name

    def get_friend(self, passport_id):
        """Return friend in this group

        Keyword arguments:
            passport_id -- string representing friend's passport ID
        """
        if self.friends.has_key(passport_id):
            return self.friends[passport_id]
        else:
            return None

    def get_friends(self):
        """Return all friends in this group"""
        return self.friends.values()

    def get_online_friends(self):
        """Return all online friends in this group"""
        return filter(lambda f: f.state != States.OFFLINE, self.friends)

class Friend:
    """A friend in the friend list"""
    def __init__(self, passport_id, display_name,
        state = States.OFFLINE, groups = None):

        self.passport_id = passport_id
        self.display_name = display_name
        self.state = state
        self.groups = {}

        if groups == None:
            groups = ()
        for g in groups:
            self.add_to_group(g)

    def get_passport_id(self):
        """Return friend's passport ID"""
        return self.passport_id

    def get_display_name(self):
        """Return friend's display name"""
        return self.display_name

    def get_state(self):
        """Return friend's presence state"""
        return self.state

    def get_groups(self):
        """Return msnp.Group instances representing friend's groups"""
        return self.groups.values()

    def add_to_group(self, group):
        """Add friend to given group"""
        group.friends[self.passport_id] = self
        self.groups[group.id] = group

    def remove_from_group(self, group):
        """Remove friend from given group"""
        del group.friends[self.passport_id]
        del self.groups[group.id]

class FriendList:
    """Friend list
    
    Includes groups, all types of lists (forward, reverse, allow, block), and
    some other book-keeping information.  msnp.Session.friend_list is an
    instance of this class.
    """
    def __init__(self):
        self.ver = 0
        self.groups = {}
        self.lists = {
            Lists.FORWARD: {},
            Lists.REVERSE: {},
            Lists.ALLOW: {},
            Lists.BLOCK: {},
        }
        self.privacy_mode = PrivacyModes.BLOCK
        self.notify_on_add_ = True
        self.dirty = True
        self.updated = time()
        self.temp_iln = {}

    def is_dirty(self):
        """If this list is dirty (needs update from server)"""
        return self.dirty

    def get_group(self, id):
        """Return msnp.Group instance corresponding to the given group ID"""
        if self.groups.has_key(id):
            return self.groups[id]
        else:
            return None

    def get_groups(self):
        """Return all groups in this list"""
        return self.groups.values()

    def get_friend(self, passport_id, list_ = Lists.FORWARD):
        """Return msnp.Friend instance for given passport ID and list type"""
        if self.lists[list_].has_key(passport_id):
            return self.lists[list_][passport_id]
        else:
            return None

    def get_friends(self, list_ = Lists.FORWARD):
        """Return all friends from given list type"""
        return self.lists[list_].values()

    def size(list_):
        """Return size (no. of members) of given list"""
        return len(self.lists[list_])

    def get_online_friends(self):
        """Return all online friends from forward list"""
        return filter(lambda f: f.state != States.OFFLINE,
            self.lists[Lists.FORWARD])

    def get_privacy_mode(self):
        """Return privacy mode (see msnp.PrivacyModes)"""
        return self.privacy_mode

    def notify_on_add(self):
        """Whether or not to be notified on being added to someone's list"""
        return self.notify_on_add_

    def last_updated(self):
        """Last updated timestamp, in seconds since the epoch"""
        return self.updated

# vim: set ts=4 sw=4 et tw=79 :




# command.py -- Command class, and subclasses
#
# Copyright (C) 2003 Manish Jethani (manish_jethani AT yahoo.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import md5
import email
import email.Message
import email.Charset

from string import split, join
from binascii import hexlify, unhexlify

from codec import url_codec

class Command:
    def __init__(self, cmd = None, trn = None, args = None):
        self.cmd = cmd
        self.trn = trn
        self.args = args

    def parse(self, str):
        s = split(str)
        pos = 0
        if s == None or len(s) <= 0: return
        self.cmd, pos = s[pos], pos + 1
        if len(s) <= pos: return
        if self.cmd in ('NLN', 'FLN',
            'GTC', 'BLP', 'PRP', 'LSG', 'LST',
            'BPR'):
            self.trn = 0
        else:
            self.trn, pos = int(s[pos]), pos + 1
        if len(s) <= pos: return
        self.args, pos = tuple(s[pos:]), pos + 1

    def __str__(self):
        cmd_trn = join((self.cmd, str(self.trn)))
        if self.args != None:
            args = join(self.args)
        else:
            args = ''
        return join((cmd_trn, args))

    def send(self, conn):
        conn.send_data_line(str(self))

class Msg(Command):
    def __init__(self):
        Command.__init__(self, 'MSG')
        self.passport_id = None
        self.display_name = None
        self.msg_len = -1
        self.msg_buf = None

    def parse(self, str):
        s = split(str)
        self.cmd = s[0]
        self.passport_id = s[1]
        self.display_name = url_codec.decode(s[2])
        self.msg_len = int(s[3])

    # for fetching the message part, after the command's '\r\n'
    def receive(self, conn):
        self.msg_buf = conn.receive_data(self.msg_len)

    def send(self, conn):
        conn.send_data_line(str(self))
        conn.send_data_all(self.msg_buf)

class Png(Command):
    def __init__(self):
        self.cmd = 'PNG'
        self.trn = 0
        self.args = None

    def parse(self, str):
        pass

    def __str__(self):
        return self.cmd

class Qry(Command):  # response to server's CHL (challenge)
    def __init__(self, trn, hash):
        Command.__init__(self, 'QRY', trn, ('msmsgs@msnmsgr.com 32',))
        self.hash = hash

    def parse(self, str):
        pass

    def __str__(self):
        s = Command.__str__(self)
        m = md5.new()
        m.update(self.hash)
        m.update('Q1P7W2E4J9R8U3S5')
        digest = hexlify(m.digest())
        return s + '\n' + digest

    def send(self, conn):
        conn.send_data_all(str(self))

# vim: set ts=4 sw=4 et tw=79 :




# chat.py -- Chat, ChatCallbacks classes
#
# Copyright (C) 2003 Manish Jethani (manish_jethani AT yahoo.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import select

import email
import email.Message
import email.Charset

from string import split, join

from session import _Session
from command import Command, Msg
from codec import url_codec

import protocol

class ChatCallbacks:
    """Callback interface for MSN chat

    The client must implement some or all of these methods to receive
    notifications on chat events.  The value of Chat.callbacks must be set
    after receiving the Chat instance in SessionCallbacks.chat_started
    """

    def friend_joined(self, passport_id, display_name):
        """Friend has joined the chat

        Keyword arguments:
            passport_id -- string representing friend's passport ID
            display_name -- friend's display name
        """

    def friend_left(self, passport_id):
        """Friend has left the chat

        Keyword arguments:
            passport_id -- string representing friend's passport ID
        """

    def message_received(self, passport_id, display_name, text, charset):
        """Message received

        Keyword arguments:
            passport_id -- string representing friend's passport ID
            display_name -- friend's display name
            text -- message text
            charset -- character set of text (usu. utf-8)
        """

    def typing_received(self, passport_id, display_name):
        """Friend is typing

        Keyword arguments:
            passport_id -- string representing friend's passport ID
            display_name -- friend's display name
        """

class Chat(_Session):
    """MSN chat conversation

    When a conversation is started, an instance of Chat is created and passed
    on to the SessionCallbacks.chat_started method.
    """
    def __init__(self, session, server, hash, passport_id, display_name,
        session_id = None, invitee = None):

        _Session.__init__(self, ChatCallbacks())

        self.session = session
        self.hash = hash
        self.passport_id = passport_id
        self.display_name = display_name
        self.session_id = session_id
        self.transaction_id = 1
        self.initial_members = []

        self.http_proxy = session.http_proxy
        conn = self.conn = self._connect(server)

        if passport_id != session.passport_id:  # invited to chat
            ans = Command('ANS', self.transaction_id,
                (session.passport_id, hash, session_id))
            self._send_cmd(ans, conn)

            while 1:
                resp = self._receive_cmd(conn)
                if resp.cmd == 'ANS':
                    break
                elif resp.cmd != 'IRO':
                    raise Error(int(resp.cmd), protocol.errors[resp.cmd])
                self.initial_members.append((resp.args[2],
                    url_codec.decode(resp.args[3])))

        else:  # hosting chat
            usr = Command('USR', self.transaction_id,
                (passport_id, hash))
            resp = self._sync_command(usr, conn)
            if resp.cmd != 'USR':
                raise Error(int(resp.cmd), protocol.errors[resp.cmd])

            cal = Command('CAL', self.transaction_id, (invitee,))
            resp = self._sync_command(cal, conn)
            if resp.cmd != 'CAL':
                raise Error(int(resp.cmd), protocol.errors[resp.cmd])
            self.session_id = resp.args[1]

    def leave(self):
        """Leave a chat conversation

        This should be the last method called on a Chat instance.
        """
        del self.session.active_chats[self.session_id]
        self.process()
        self.conn.break_()
        self.conn = None

    def __send_mime_message(self, mime_message, flag):
        msg = Msg()
        msg.trn = self.transaction_id

        msg.msg_buf = ''
        for hdr in mime_message.items():
            msg.msg_buf = msg.msg_buf + join(hdr, ': ') + '\r\n'
        msg.msg_buf = msg.msg_buf + '\r\n'
        if mime_message.get_payload() != None:
            msg.msg_buf = msg.msg_buf + mime_message.get_payload()

        msg.args = (flag, str(len(msg.msg_buf)))

        self._async_command(msg)
        self.process()

    def send_message(self, text, charset = 'utf-8'):
        """Send message

        Keyword arguments:
            text -- message text
            charset -- character set of text (default utf-8)
        """
        mime_message = email.Message.Message()
        mime_message['MIME-Version'] = '1.0'
        mime_message['Content-Type'] = 'text/plain; charset=' + charset
        mime_message.set_payload(text)
        self.__send_mime_message(mime_message, 'N')

    def send_typing(self):
        """Send typing notification"""
        mime_message = email.Message.Message()
        mime_message['MIME-Version'] = '1.0'
        mime_message['Content-Type'] = 'text/x-msmsgscontrol'
        mime_message['TypingUser'] = self.session.passport_id
        self.__send_mime_message(mime_message, 'U')

    def process(self):
        """Process events

        This method must be called periodically, preferably in the client
        application's main loop.
        """
        while 1:
            if self.conn == None:
                break
            fd = self.conn.socket.fileno()
            r = select.select([fd], [], [], 0)
            if len(r[0]) > 0:
                buf = self.conn.receive_data_line()
                if buf == None:  # connection closed?!
                    break
                self.__process_command_buf(buf)
            elif len(self.send_queue) > 0:
                cmd = self.send_queue.pop(0)
                cmd.send(self.conn)
            else:
                break

    def __process_command_buf(self, buf):
        cmd = buf[:3]
        if cmd == 'MSG':
            self.__process_msg(buf)
        elif cmd == 'JOI':
            self.__process_joi(buf)
        elif cmd == 'BYE':
            self.__process_bye(buf)
        # TODO error handling

    def __process_msg(self, buf):
        msg = Msg()
        msg.parse(buf)
        msg.receive(self.conn)
        mime_message = email.message_from_string(msg.msg_buf)

        if mime_message.get_content_type() == 'text/plain':
            self.callbacks.message_received(msg.passport_id,
                msg.display_name,
                mime_message.get_payload(),
                mime_message.get_content_charset())

        elif mime_message.get_content_type() == 'text/x-msmsgscontrol':
            self.callbacks.typing_received(msg.passport_id,
            msg.display_name)

    def __process_joi(self, buf):
        joi = split(buf)
        self.callbacks.friend_joined(joi[1], url_codec.decode(joi[2]))

    def __process_bye(self, buf):
        bye = split(buf)
        self.callbacks.friend_left(bye[1])

# vim: set ts=4 sw=4 et tw=79 :

